//////////////////////////////////////////////////////////////////////////////
//
//                     (C) Copyright 2018 by Autodesk, Inc.
//
// The information contained herein is confidential, proprietary to Autodesk,
// Inc., and considered a trade secret as defined in section 499C of the
// penal code of the State of California.  Use of this information by anyone
// other than authorized employees of Autodesk, Inc. is granted only under a
// written non-disclosure agreement, expressly prescribing the scope and
// manner of such use.
//
//        AUTODESK PROVIDES THIS PROGRAM "AS IS" AND WITH ALL FAULTS.
//        AUTODESK SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTY OF
//        MERCHANTABILITY OR FITNESS FOR A PARTICULAR USE.  AUTODESK, INC.
//        DOES NOT WARRANT THAT THE OPERATION OF THE PROGRAM WILL BE
//        UNINTERRUPTED OR ERROR FREE.
//
//////////////////////////////////////////////////////////////////////////////

#include <vector>

#ifdef __clang__
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wdocumentation"
#pragma clang diagnostic ignored "-Wmissing-noreturn"
#pragma clang diagnostic ignored "-Wdeprecated"
#pragma clang diagnostic ignored "-Wused-but-marked-unused"
#endif
#include <Eigen/Eigenvalues>
#ifdef __clang__
#pragma clang diagnostic pop
#endif

#include <common/RCLinearAlgebra.h>
#include <common/RCVector.h>

using namespace Autodesk::RealityComputing::Foundation;

class RCEigenSolverSymmetric3x3::Impl
{
public:
    Impl(const double xx, const double xy, const double xz, const double yy, const double yz, const double zz);
    RCVector3d getEigenvalues() const
    {
        return eigenvalues;
    }
    RCVector3d getEigenvector(const size_t index) const
    {
        if (index < eigenvectors.size())
            return eigenvectors[index];
        else
            return RCVector3d();
    }

private:
    RCVector3d eigenvalues;
    std::vector<RCVector3d> eigenvectors;
};

RCEigenSolverSymmetric3x3::Impl::Impl(const double xx, const double xy, const double xz, const double yy, const double yz, const double zz)
{
    Eigen::Matrix3d matrix;
    matrix(0, 0) = xx;
    matrix(0, 1) = xy;
    matrix(0, 2) = xz;
    matrix(1, 1) = yy;
    matrix(1, 2) = yz;
    matrix(2, 2) = zz;
    // symmetry
    matrix(1, 0) = xy;
    matrix(2, 0) = xz;
    matrix(2, 1) = yz;

    // solve the eigensystem
    Eigen::SelfAdjointEigenSolver<Eigen::Matrix3d> eigensystem(matrix);

    // copy the eigenvalues and eigenvectors
    eigenvalues = RCVector3d(eigensystem.eigenvalues().x(), eigensystem.eigenvalues().y(), eigensystem.eigenvalues().z());
    eigenvectors.reserve(3);
    for (size_t i = 0; i < 3; i++)
        eigenvectors.push_back(RCVector3d(eigensystem.eigenvectors().col(i).x(), eigensystem.eigenvectors().col(i).y(), eigensystem.eigenvectors().col(i).z()));
};

RCEigenSolverSymmetric3x3::RCEigenSolverSymmetric3x3(const double xx, const double xy, const double xz, const double yy, const double yz, const double zz)
    : mImpl(new Impl(xx, xy, xz, yy, yz, zz))
{
}

RCEigenSolverSymmetric3x3::~RCEigenSolverSymmetric3x3()
{
    if (mImpl)
        delete mImpl;
}

RCVector3d RCEigenSolverSymmetric3x3::getEigenvalues() const
{
    return mImpl->getEigenvalues();
}

RCVector3d RCEigenSolverSymmetric3x3::getEigenvector(const size_t index) const
{
    return mImpl->getEigenvector(index);
}
